#include "plane.h"
#include "vertex.h"
#include "sphere.h"
#include "ray.h"
#include "light.h"
#include "ppmconverter.h"
#include "pixel2world.h"
#include "parsecommandline.h"
#include "raytrace.h"
#include "intersection.h"
#include "phong.h"
#include "povray.h"

#include <stdlib.h>
#include <stdio.h>
#include <vector>
#include <string>
#include <math.h>

using namespace std;

void generateStartingRays(void);

int scene_width = 0;
int scene_height = 0;
int stages = 3;
string in_file;
string out_file;

vector<Sphere*> spheres;
vector<Plane*> planes;
vector<Light*> lights;
vector<Pixel*> pixels;
vector<Ray*> startingRays;

double cx, cy, cz;
double lax, lay, laz;

int main(int argc, char **argv)
{ 
  cx = 0;
  cy = 0;
  cz = 4;
  
  parse_commandLine(argc, argv);
  generateStartingRays();
  
  int size = startingRays.size();
  /*spheres.push_back(new Sphere(-2, 0, -2, 1, 1, .9, .7, 1, 1, 1, .125, .5, .5, 5.0, .5));
  spheres.push_back(new Sphere(0, 0, -2, 1, 0, 0, 0, 1, 1, 1, .125, .5, .5, 5.0, .2));
  spheres.push_back(new Sphere(2, 0, -2, 1, 1, 1, 1, 1, 1, 1, .125, .5, .5, 5.0, .7));
  //spheres.push_back(new Sphere(0, .8, -2, 1, 0, 0, 1, 1, 1, 1, .125, .5, .4, 5.0, .4));
  planes.push_back(new Plane(0, 1, 0, 3.2, 1, 0, .5, 1, 1, 1, .125, .5, .5, 5.0, .5));
  //lights.push_back(new Light(-1, 0, -4, 1, 1, 1));
  lights.push_back(new Light(0, 8, 0, 1, 1, 1));*/
  
  POVRAYParser parser(in_file.c_str());
  
  parser.parseFile();
  
  cx = parser.camx;
  cy = parser.camy;
  cz = parser.camz;
  
  printf("CAMERA:\n");
  printf("<%lf, %lf, %lf>\n", cx, cy, cz);
  
  for (int i = 0; i < lights.size(); i++)
  {
    printf("LIGHT:\n");
    printf("<%lf, %lf, %lf> %lf, %lf, %lf>\n", lights[i]->pos.x, lights[i]->pos.y, lights[i]->pos.z,
                                               lights[i]->color.r, lights[i]->color.g, lights[i]->color.b);
  }
  
  for (int i = 0; i < spheres.size(); i++)
  {
    printf("SPHERE:\n");
    printf("<%lf, %lf, %lf, %lf>\n", spheres[i]->center->x, spheres[i]->center->y, spheres[i]->center->z, spheres[i]->radius);
    printf("<%lf, %lf, %lf>\n", spheres[i]->material->diffuse.r, spheres[i]->material->diffuse.g,  spheres[i]->material->diffuse.b);
    printf("<%lf, %lf, %lf>\n", spheres[i]->material->specular.r, spheres[i]->material->specular.g,  spheres[i]->material->specular.b);
    printf("%lf, %lf, %lf, ", spheres[i]->material->ambient_coefficient, spheres[i]->material->diffuse_coefficient, spheres[i]->material->specular_coefficient);
    printf("%lf, %lf, %lf\n", spheres[i]->material->roughness, spheres[i]->material->reflectivity);
  }
  
  for (int i = 0; i < planes.size(); i++)
  {
    printf("PLANE:\n");
    printf("<%lf, %lf, %lf> ", planes[i]->normal->dx, planes[i]->normal->dy, planes[i]->normal->dz);
    printf("<%lf, %lf, %lf>\n", planes[i]->point->x, planes[i]->point->y, planes[i]->point->z);
    printf("<%lf, %lf, %lf>\n", planes[i]->material->diffuse.r, planes[i]->material->diffuse.g, planes[i]->material->diffuse.b);
    printf("<%lf, %lf, %lf>\n", planes[i]->material->specular.r, planes[i]->material->specular.g,  planes[i]->material->specular.b);
    printf("%lf, %lf, %lf, ", planes[i]->material->ambient_coefficient, planes[i]->material->diffuse_coefficient, planes[i]->material->specular_coefficient);
    printf("%lf, %lf, %lf\n", planes[i]->material->roughness, planes[i]->material->reflectivity);
  }
  
  //exit(0);
  
  
  Pixel *pixel;
  
  for (int i = 0; i < size; i++)
  {
    pixel = raytrace(startingRays[i]);
    
    pixel->red = pixel->red > 1 ? 1 : pixel->red;
    pixel->green = pixel->green > 1 ? 1 : pixel->green;
    pixel->blue = pixel->blue > 1 ? 1 : pixel->blue;

    pixel->red = pixel->red < 0 ? 0 : pixel->red;
    pixel->green = pixel->green < 0 ? 0 : pixel->green;
    pixel->blue = pixel->blue < 0 ? 0 : pixel->blue;
    
    pixels.push_back(pixel);
  }

  if (pixels.size() == 0)
    exit(100);
  

  PPMConverter ppm(out_file.c_str());  
  ppm.writeFile(&pixels);
  ppm.finish();
  
  for (int i = 0; i < startingRays.size(); i++)
  {
    delete startingRays[i];
  }
  
  return 0;
}

void generateStartingRays()
{
  double camx = pixel2world_x(scene_width/2);
  double camy = pixel2world_y(scene_height/2);
  double camz = 100;
  double fov = 255.0;

  for (int y = 0; y < scene_height; y++)
  {
    for (int x = 0; x < scene_width; x++)
    {
      double xn = pixel2world_x(x);
      double yn = pixel2world_y(y);
    
      Ray *add = new Ray((double)0, (double)0, (double)4.0, (double)(xn-cx), (double)(yn-cy), 1); //cx, cy, cz, dx, dy, fov
      add->normalize();
      startingRays.push_back(add);
    }
  }
}
