#include <cstdio>
#include <cmath>

using namespace std;

#include "phong.h"
#include "ray.h"
#include "light.h"
#include "sphere.h"
#include "plane.h"

void getPhongColors(int objType, void *object, Ray *ray, double rgb[3], double ix, double iy, double iz)
{
  rgb[0] = 0;
  rgb[1] = 0;
  rgb[2] = 0;
  
  if (objType == 0)
  {
    Sphere *obj = static_cast<Sphere*>(object);
    int size = lights.size();
    
    for (int i = 0; i < size; i++)
    {
      Light *l = lights[i];
      
      Ray lv(ix, iy, iz, l->pos.x + ix, l->pos.y + iy, l->pos.z + iz);
      lv.normalize();
      
      //lv.dx = 0;
      //lv.dy = -3.0/5;
      //lv.dz = -4.0/5;
      
      //double nx = -(obj->center->x - ix);
      //double ny = -(obj->center->y - iy);
      //double nz = -(obj->center->z - iz);
      
      double nx = (ix - obj->center->x);
      double ny = (iy - obj->center->y);
      double nz = (iz - obj->center->z);
      
      Ray snormal(obj->center->x, obj->center->y, obj->center->z, nx, ny, nz);
      snormal.normalize();
      
      double ndotl = lv ^ snormal;
      
      ndotl = ndotl < 0 ? 0 : ndotl;
      
      double diffuse_r = ndotl * obj->material->diffuse_coefficient * obj->material->diffuse.r * l->color.r;
      double diffuse_g = ndotl * obj->material->diffuse_coefficient * obj->material->diffuse.g * l->color.g;
      double diffuse_b = ndotl * obj->material->diffuse_coefficient * obj->material->diffuse.b * l->color.b;
        
      //double ndotl = snormal ^ lv;
      
      double ndotl_2 = 2 * ndotl;
      
      //snormal * 2;
      //snormal.normalize();
      
      snormal * ndotl_2;
      //snormal.normalize();
        
      Ray *refl = snormal - lv;      
      refl->normalize();
      
      Ray view(ix, iy, iz, ray->ox - ix, ray->oy - iy, ray->oz - iz);
      view.normalize();
      
      double vdotr = *refl ^ view;
      
      vdotr = vdotr < 0 ? 0 : vdotr;
      
      //double vdotr = *refl ^ *ray;
      double vdotr_ruff = pow(vdotr, obj->material->roughness);
      
      double specular_r = vdotr_ruff * obj->material->specular_coefficient /* obj->material->specular.r*/ * l->color.r;
      double specular_g = vdotr_ruff * obj->material->specular_coefficient /* obj->material->specular.g*/ * l->color.g;
      double specular_b = vdotr_ruff * obj->material->specular_coefficient /* obj->material->specular.b*/ * l->color.b;
      
      double ambient_r = obj->material->diffuse.r * obj->material->ambient_coefficient * l->color.r;
      double ambient_g = obj->material->diffuse.g * obj->material->ambient_coefficient * l->color.g;
      double ambient_b = obj->material->diffuse.b * obj->material->ambient_coefficient * l->color.b;

      printf("a: <%lf, %lf, %lf>\n", ambient_r, ambient_g, ambient_b);
      printf("d: <%lf, %lf, %lf>\n", diffuse_r, diffuse_g, diffuse_b);
      printf("s: <%lf, %lf, %lf>\n\n", specular_r, specular_g, specular_b);
      
      rgb[0] += ambient_r + diffuse_r + specular_r;
      rgb[1] += ambient_g + diffuse_g + specular_g;
      rgb[2] += ambient_b + diffuse_b + specular_b;
    }
    
    rgb[0] = rgb[0] < 0 ? 0 : rgb[0];
    rgb[1] = rgb[1] < 0 ? 0 : rgb[1];
    rgb[2] = rgb[2] < 0 ? 0 : rgb[2];
    
    rgb[0] = rgb[0] > 1 ? 1 : rgb[0];
    rgb[1] = rgb[1] > 1 ? 1 : rgb[1];
    rgb[2] = rgb[2] > 1 ? 1 : rgb[2];
    
    //printf("<%lf, %lf,%lf>\n", rgb[0], rgb[1], rgb[2]);
  }
  else
  {
    Plane *obj = static_cast<Plane*>(object);
    int size = lights.size();
    
    for (int i = 0; i < size; i++)
    {
      Light *l = lights[i];
      
      Ray lv(ix, iy, iz, l->pos.x + ix, l->pos.y + iy, l->pos.z + iz);
      lv.normalize();
      
      //lv.dx = 0;
      //lv.dy = -3.0/5;
      //lv.dz = -4/5/5;
      
      //double nx = -(obj->center->x - ix);
      //double ny = -(obj->center->y - iy);
      //double nz = -(obj->center->z - iz);
      
      //double nx = (ix - obj->center->x);
      //double ny = (iy - obj->center->y);
      //double nz = (iz - obj->center->z);
      
      Ray snormal(obj->normal->ox, obj->normal->oy, obj->normal->oz, obj->normal->dx, obj->normal->dy, obj->normal->dz);
      snormal.normalize();
      
      double ndotl = lv ^ snormal;

      ndotl = ndotl < 0 ? 0 : ndotl;
      
      printf("NDOTL: %lf\n", ndotl);

      double diffuse_r = ndotl * obj->material->diffuse_coefficient * obj->material->diffuse.r * l->color.r;
      double diffuse_g = ndotl * obj->material->diffuse_coefficient * obj->material->diffuse.g * l->color.g;
      double diffuse_b = ndotl * obj->material->diffuse_coefficient * obj->material->diffuse.b * l->color.r;
        
      //double ndotl = snormal ^ lv;
      
      double ndotl_2 = 2 * ndotl;
      
      //snormal * 2;
      //snormal.normalize();
      
      snormal * ndotl_2;
      //snormal.normalize();
        
      Ray *refl = snormal - lv;
      refl->normalize();
      
      Ray view(ix, iy, iz, ray->ox - ix, ray->oy - iy, ray->oz - iz);
      view.normalize();
      
      double vdotr = *refl ^ view;
      
      vdotr = vdotr < 0 ? 0 : vdotr;
      
      //double vdotr = *refl ^ *ray;
      double vdotr_ruff = pow(vdotr, obj->material->roughness);
      
      double specular_r = vdotr_ruff * obj->material->specular_coefficient /** obj->material->specular.r*/ * l->color.r;
      double specular_g = vdotr_ruff * obj->material->specular_coefficient /** obj->material->specular.g*/ * l->color.g;
      double specular_b = vdotr_ruff * obj->material->specular_coefficient /** obj->material->specular.b*/ * l->color.b;
      
      double ambient_r = obj->material->diffuse.r * obj->material->ambient_coefficient * l->color.r;
      double ambient_g = obj->material->diffuse.g * obj->material->ambient_coefficient * l->color.g;
      double ambient_b = obj->material->diffuse.b * obj->material->ambient_coefficient * l->color.b;
      
      //printf("a: <%lf, %lf, %lf>\n", ambient_r, ambient_g, ambient_b); 
                   
      rgb[0] = rgb[0] + ambient_r + diffuse_r + specular_r;
      rgb[1] = rgb[1] + ambient_g + diffuse_g + specular_g;
      rgb[2] = rgb[2] + ambient_b + diffuse_b + specular_b;
    }
    
    rgb[0] = rgb[0] < 0 ? 0 : rgb[0];
    rgb[1] = rgb[1] < 0 ? 0 : rgb[1];
    rgb[2] = rgb[2] < 0 ? 0 : rgb[2];
    
    rgb[0] = rgb[0] > 1 ? 1 : rgb[0];
    rgb[1] = rgb[1] > 1 ? 1 : rgb[1];
    rgb[2] = rgb[2] > 1 ? 1 : rgb[2];
    
    //printf("<%lf, %lf,%lf>\n", rgb[0], rgb[1], rgb[2]); 
  }
}

void getPhongColors_withShadows(int objType, void *object, Ray *ray, double rgb[3], double ix, double iy, double iz)
{
  rgb[0] = 0;
  rgb[1] = 0;
  rgb[2] = 0;
  
  if (objType == 0)
  {
    Sphere *obj = static_cast<Sphere*>(object);
    int size = lights.size();
    
    for (int i = 0; i < size; i++)
    {
      Light *l = lights[i];

      /*double distance_to_light = (l->pos.x + ix)*(l->pos.x + ix) +
                                 (l->pos.y + iy)*(l->pos.y + iy) +
                                 (l->pos.z + iz)*(l->pos.z + iz);
                                 
      distance_to_light = sqrt(distance_to_light);*/
      
      Ray lv(ix, iy, iz, l->pos.x + ix, l->pos.y + iy, l->pos.z + iz);

      double distance_to_light = lv.mag;

      lv.normalize();
      
      double ox = ix + (l->pos.x + ix) * .1;
      double oy = iy + (l->pos.y + iy) * .1;
      double oz = iz + (l->pos.z + iz) * .1;
      
      printf("LIGHT: (%lf, %lf, %lf)\n", l->pos.x, l->pos.y, l->pos.z);
      printf("INTERSECTION: (%lf, %lf, %lf)\n", ix, iy, iz);
      
      Ray lv2(ix, iy, iz, l->pos.x + ix, l->pos.y + iy, -l->pos.z + iz);
      lv2.normalize();
      
      if (!isObstructed(&lv, ix, iy, iz, distance_to_light, l))
      {
        printf("------------------- SPHERE IS OBSTRUCTED (%lf, %lf, %lf)-------------------\n", ix, iy, iz);
        double ambient_r = obj->material->diffuse.r * obj->material->ambient_coefficient * l->color.r;
        double ambient_g = obj->material->diffuse.g * obj->material->ambient_coefficient * l->color.g;
        double ambient_b = obj->material->diffuse.b * obj->material->ambient_coefficient * l->color.b;      
           
        rgb[0] = rgb[0] + ambient_r;
        rgb[1] = rgb[1] + ambient_g;
        rgb[2] = rgb[2] + ambient_b;
      }
      else
      {
      //lv.dx = 0;
      //lv.dy = -3.0/5;
      //lv.dz = -4.0/5;
      
      //double nx = -(obj->center->x - ix);
      //double ny = -(obj->center->y - iy);
      //double nz = -(obj->center->z - iz);
      
      double nx = (ix - obj->center->x);
      double ny = (iy - obj->center->y);
      double nz = (iz - obj->center->z);
      
      Ray snormal(obj->center->x, obj->center->y, obj->center->z, nx, ny, nz);
      snormal.normalize();
      
      double ndotl = lv ^ snormal;
      
      ndotl = ndotl < 0 ? 0 : ndotl;
      
      double diffuse_r = ndotl * obj->material->diffuse_coefficient * obj->material->diffuse.r * l->color.r;
      double diffuse_g = ndotl * obj->material->diffuse_coefficient * obj->material->diffuse.g * l->color.g;
      double diffuse_b = ndotl * obj->material->diffuse_coefficient * obj->material->diffuse.b * l->color.b;
        
      //double ndotl = snormal ^ lv;
      
      double ndotl_2 = 2 * ndotl;
      
      //snormal * 2;
      //snormal.normalize();
      
      snormal * ndotl_2;
      //snormal.normalize();
        
      Ray *refl = snormal - lv;      
      refl->normalize();
      
      Ray view(ix, iy, iz, ray->ox - ix, ray->oy - iy, ray->oz - iz);
      view.normalize();
      
      double vdotr = *refl ^ view;
      
      vdotr = vdotr < 0 ? 0 : vdotr;
      
      //double vdotr = *refl ^ *ray;
      double vdotr_ruff = pow(vdotr, obj->material->roughness);
      
      double specular_r = vdotr_ruff * obj->material->specular_coefficient /* obj->material->specular.r*/ * l->color.r;
      double specular_g = vdotr_ruff * obj->material->specular_coefficient /* obj->material->specular.g*/ * l->color.g;
      double specular_b = vdotr_ruff * obj->material->specular_coefficient /* obj->material->specular.b*/ * l->color.b;
      
      double ambient_r = obj->material->diffuse.r * obj->material->ambient_coefficient * l->color.r;
      double ambient_g = obj->material->diffuse.g * obj->material->ambient_coefficient * l->color.g;
      double ambient_b = obj->material->diffuse.b * obj->material->ambient_coefficient * l->color.b;

      //printf("a: <%lf, %lf, %lf>\n", ambient_r, ambient_g, ambient_b);
      //printf("d: <%lf, %lf, %lf>\n", diffuse_r, diffuse_g, diffuse_b);
     // printf("s: <%lf, %lf, %lf>\n\n", specular_r, specular_g, specular_b);
      
      rgb[0] += ambient_r + diffuse_r + specular_r;
      rgb[1] += ambient_g + diffuse_g + specular_g;
      rgb[2] += ambient_b + diffuse_b + specular_b;
      }
    }
    
    rgb[0] = rgb[0] < 0 ? 0 : rgb[0];
    rgb[1] = rgb[1] < 0 ? 0 : rgb[1];
    rgb[2] = rgb[2] < 0 ? 0 : rgb[2];
    
    rgb[0] = rgb[0] > 1 ? 1 : rgb[0];
    rgb[1] = rgb[1] > 1 ? 1 : rgb[1];
    rgb[2] = rgb[2] > 1 ? 1 : rgb[2];
    
    //printf("<%lf, %lf,%lf>\n", rgb[0], rgb[1], rgb[2]);
  }
  else
  {
    Plane *obj = static_cast<Plane*>(object);
    int size = lights.size();
    
    for (int i = 0; i < size; i++)
    {
      Light *l = lights[i];
      
      /*double distance_to_light = (l->pos.x + ix)*(l->pos.x + ix) +
                                 (l->pos.y + iy)*(l->pos.y + iy) +
                                 (l->pos.z + iz)*(l->pos.z + iz);
      
      distance_to_light = sqrt(distance_to_light); */
      
      Ray lv(ix, iy, iz, l->pos.x + ix, l->pos.y + iy, l->pos.z + iz);

      double distance_to_light = lv.mag;
   
      lv.normalize();
      
      double ox = ix + (l->pos.x + ix) * .0001;
      double oy = iy + (l->pos.y + iy) * .0001;
      double oz = iz + (l->pos.z + iz) * .0001;
      
      Ray lv2(ox, oy, oz, l->pos.x + ix, l->pos.y + iy, l->pos.z + iz);
      lv2.normalize();
      
      if (isObstructed(&lv, ox, oy, oz, distance_to_light, l))
      {
        printf("------------------- PLANE IS OBSTRUCTED (%lf, %lf, %lf)-------------------\n", ix, iy, iz);
        double ambient_r = obj->material->diffuse.r * obj->material->ambient_coefficient * l->color.r;
        double ambient_g = obj->material->diffuse.g * obj->material->ambient_coefficient * l->color.g;
        double ambient_b = obj->material->diffuse.b * obj->material->ambient_coefficient * l->color.b;      
           
        rgb[0] = rgb[0] + ambient_r;
        rgb[1] = rgb[1] + ambient_g;
        rgb[2] = rgb[2] + ambient_b;
        
        continue;
      }
      else
      {
      //lv.dx = 0;
      //lv.dy = -3.0/5;
      //lv.dz = -4/5/5;
      
      //double nx = -(obj->center->x - ix);
      //double ny = -(obj->center->y - iy);
      //double nz = -(obj->center->z - iz);
      
      //double nx = (ix - obj->center->x);
      //double ny = (iy - obj->center->y);
      //double nz = (iz - obj->center->z);
      
      Ray snormal(obj->normal->ox, obj->normal->oy, obj->normal->oz, obj->normal->dx, obj->normal->dy, obj->normal->dz);
      snormal.normalize();
      
      double ndotl = lv ^ snormal;

      ndotl = ndotl < 0 ? 0 : ndotl;
      
      //printf("NDOTL: %lf\n", ndotl);

      double diffuse_r = ndotl * obj->material->diffuse_coefficient * obj->material->diffuse.r * l->color.r;
      double diffuse_g = ndotl * obj->material->diffuse_coefficient * obj->material->diffuse.g * l->color.g;
      double diffuse_b = ndotl * obj->material->diffuse_coefficient * obj->material->diffuse.b * l->color.r;
        
      //double ndotl = snormal ^ lv;
      
      double ndotl_2 = 2 * ndotl;
      
      //snormal * 2;
      //snormal.normalize();
      
      snormal * ndotl_2;
      //snormal.normalize();
        
      Ray *refl = snormal - lv;
      refl->normalize();
      
      Ray view(ix, iy, iz, ray->ox - ix, ray->oy - iy, ray->oz - iz);
      view.normalize();
      
      double vdotr = *refl ^ view;
      
      vdotr = vdotr < 0 ? 0 : vdotr;
      
      //double vdotr = *refl ^ *ray;
      double vdotr_ruff = pow(vdotr, obj->material->roughness);
      
      double specular_r = vdotr_ruff * obj->material->specular_coefficient /** obj->material->specular.r*/ * l->color.r;
      double specular_g = vdotr_ruff * obj->material->specular_coefficient /** obj->material->specular.g*/ * l->color.g;
      double specular_b = vdotr_ruff * obj->material->specular_coefficient /** obj->material->specular.b*/ * l->color.b;
      
      double ambient_r = obj->material->diffuse.r * obj->material->ambient_coefficient * l->color.r;
      double ambient_g = obj->material->diffuse.g * obj->material->ambient_coefficient * l->color.g;
      double ambient_b = obj->material->diffuse.b * obj->material->ambient_coefficient * l->color.b;
      
      //printf("a: <%lf, %lf, %lf>\n", ambient_r, ambient_g, ambient_b); 
                   
      rgb[0] = rgb[0] + ambient_r + diffuse_r + specular_r;
      rgb[1] = rgb[1] + ambient_g + diffuse_g + specular_g;
      rgb[2] = rgb[2] + ambient_b + diffuse_b + specular_b;
      }
    }
    
    rgb[0] = rgb[0] < 0 ? 0 : rgb[0];
    rgb[1] = rgb[1] < 0 ? 0 : rgb[1];
    rgb[2] = rgb[2] < 0 ? 0 : rgb[2];
    
    rgb[0] = rgb[0] > 1 ? 1 : rgb[0];
    rgb[1] = rgb[1] > 1 ? 1 : rgb[1];
    rgb[2] = rgb[2] > 1 ? 1 : rgb[2];
    
    //printf("<%lf, %lf,%lf>\n", rgb[0], rgb[1], rgb[2]); 
  }
}

bool isObstructed(Ray* ray, double ix, double iy, double iz, double distance_to_light, Light *l)
{
  Vertex *intersection = NULL;
  
  for (int j = 0, num_spheres = spheres.size(); j < num_spheres; j++)
  {
    printf("get intersection...\n");
  
    intersection = spheres[j]->intersects(ray);

    if (intersection != NULL)
    {
      double distance_to_intersection = (intersection->x + ix)*(intersection->x + ix) +
                                        (intersection->y + iy)*(intersection->y + iy) +
                                        (intersection->z + iz)*(intersection->z + iz);
  
      distance_to_intersection = sqrt(distance_to_intersection);
      
      Ray r1(ix, iy, iz, intersection->x + ix, intersection->y + iy, intersection->z + iz);
      
      printf("S - dti: %lf, dil: %lf\n", distance_to_intersection, distance_to_light);
      printf("  - r1 mag: %lf\n", r1.mag);
      
      delete intersection;
      
      if (distance_to_intersection < distance_to_light)
        return true;
    }
  }
  
  for (int j = 0, num_planes = planes.size(); j < num_planes; j++)
  {
    intersection = planes[j]->intersects(ray);

    if (intersection != NULL)
    {
      double distance_to_intersection = (intersection->x + ix)*(intersection->x + ix) +
                                        (intersection->y + iy)*(intersection->y + iy) +
                                        (intersection->z + iz)*(intersection->z + iz);
  
      distance_to_intersection = sqrt(distance_to_intersection);
      
      printf("P - dti: %lf, dil: %lf\n", distance_to_intersection, distance_to_light);
            
      delete intersection;
      
      if (distance_to_intersection < distance_to_light)
        return true;
    }
  }
  
  return false;
}
