#include "plane.h"
#include "pixel.h"
#include "ray.h"
#include "material.h"

#include <stdlib.h>
#include <cstdio>

Plane::Plane(double x, double y, double z, double _loc,
             double dr, double dg, double db,
             double sr, double sg, double sb,
             double ac, double dc, double sc, double ruff, double ref)
{  
  normal = new Ray(0, 0, 0, x, y, z);
  normal->normalize();
  
  material = new Material(dr, dg, db, sr, sg, sb, ac, dc, sc, ruff, ref);

  double px = 0, py = 0, pz = 0;

  if (x == 1)
    px = _loc;
  else if (y == 1)
    py = _loc;
  else
    pz = _loc;
  
  point = new Vertex(px, py, pz);
  
  loc = _loc;
}

Plane::~Plane()
{
  delete normal;
  delete point;
  delete material;
}

Vertex *Plane::intersects(Ray *ray)
{
  Ray orig(0, 0, 0, ray->ox, ray->oy, ray->oz);
  //Ray normal(point->x, point->y, point->z, this->normal->dx, normal->dy, normal->dz);
  
  double norm_dot_dir  = *normal ^ *ray; //Vd
  
  if (norm_dot_dir <= 0)
  {
    return NULL;
  }
  
  double vd = norm_dot_dir;
  double norm_dot_orig = *normal ^ orig;  
  double v0 = -(norm_dot_orig + loc);
  
  vd = vd > 0 ? -vd : vd;
  
  double t = v0 / vd;
  
  if (t < 0)
  {
    return NULL;
  }
  
  double x = ray->ox + (ray->dx*t);
  double y = ray->oy + (ray->dy*t);
  double z = ray->oz + (ray->dz*t);
  
  return new Vertex(x, y, z);
}
