#include <cstdlib>
#include <cstdio>
#include <cstring>

#include "povray.h"

POVRAYParser::POVRAYParser(const char *fname)
{
  if ((file = fopen(fname, "r")) == NULL)
  {
    printf("unable to open '%s' for reading\n", fname);
    exit(1);
  }
}

int POVRAYParser::parseFile()
{
  char line[100];
  
  while (!feof(file))
  {
    //get a line
    fgets(line, 100, file);
    
    if (strncmp(line, "camera", 6) == 0)
    {
      printf("parse camera\n");
      
      //flush out the '{'
      fgets(line, 100, file);
      //get location
      fgets(line, 100, file);
      
      char *tok = strtok(line, "\n\t <>,");
      
      tok = strtok(NULL, "\n\t <>,");
      camx = strtod(tok, NULL);
      
      tok = strtok(NULL, "\n\t <>,"); 
      camy = strtod(tok, NULL);
      
      tok = strtok(NULL, "\n\t <>,");
      camz = strtod(tok, NULL);
      
      printf("camera: %lf, %lf, %lf\n", camx, camy, camz);
      
      //get look at
      fgets(line, 100, file);
      
      tok = strtok(line, "\n\t <>,");
      
      tok = strtok(NULL, "\n\t <>,");
      lax = strtod(tok, NULL);
      
      tok = strtok(NULL, "\n\t <>,");
      lay = strtod(tok, NULL);
      
      tok = strtok(NULL, "\n\t <>,");
      laz = strtod(tok, NULL);
      
      printf("look_at: %lf, %lf, %lf\n", lax, lay, laz);
      
    }
    else if (strncmp(line, "light_source", 12) == 0)
    {
      printf("parse light_source\n");
      
      //flush out the '{'
      fgets(line, 100, file);
      //get light location
      fgets(line, 100, file);
      
      char *tok = strtok(line, "\n\t <>,");
      double lx = strtod(tok, NULL);
      
      tok = strtok(NULL, "\n\t <>,"); 
      double ly = strtod(tok, NULL);
      
      tok = strtok(NULL, "\n\t <>,");
      double lz = strtod(tok, NULL);
      
      printf("light location: %lf, %lf, %lf\n", lx, ly, lz);
      
      //get light color
      fgets(line, 100, file);
      
      tok = strtok(line, "\n\t <>,"); 
      tok = strtok(NULL, "\n\t <>,");
      
      tok = strtok(NULL, "\n\t <>,");
      double lcr = strtod(tok, NULL);
      
      tok = strtok(NULL, "\n\t <>,");
      double lcg = strtod(tok, NULL);
      
      tok = strtok(NULL, "\n\t <>,");
      double lcb = strtod(tok, NULL);
      
      printf("light color: %lf, %lf, %lf\n", lcr, lcg, lcb);
      
      lights.push_back(new Light(lx, ly, lz, lcr, lcg, lcb));
    }
    else if (strncmp(line, "sphere", 6) == 0)
    {
      printf("parse sphere\n");
      
      //flush out the '{'
      fgets(line, 100, file);
      //get light location
      fgets(line, 100, file);
      
      char *tok = strtok(line, "\n\t <>,");
      double sx = strtod(tok, NULL);
      
      tok = strtok(NULL, "\n\t <>,"); 
      double sy = strtod(tok, NULL);
      
      tok = strtok(NULL, "\n\t <>,");
      double sz = strtod(tok, NULL);
      
      tok = strtok(NULL, "\n\t <>,");
      double sr = strtod(tok, NULL);
      
      printf("sphere position/radius: %lf, %lf, %lf (%lf)\n", sx, sy, sz, sr);
      
      //get sphere diffuse color
      fgets(line, 100, file);
      
      tok = strtok(line, "\n\t <>,"); 
      tok = strtok(NULL, "\n\t <>,");
      
      tok = strtok(NULL, "\n\t <>,");
      double scr = strtod(tok, NULL);
      
      tok = strtok(NULL, "\n\t <>,");
      double scg = strtod(tok, NULL);
      
      tok = strtok(NULL, "\n\t <>,");
      double scb = strtod(tok, NULL);
      
      printf("sphere diffuse color: %lf, %lf, %lf\n", scr, scg, scb);
      
      //get sphere specular color
      fgets(line, 100, file);
      
      tok = strtok(line, "\n\t <>,"); 
      tok = strtok(NULL, "\n\t <>,");
      
      tok = strtok(NULL, "\n\t <>,");
      double scr2 = strtod(tok, NULL);
      
      tok = strtok(NULL, "\n\t <>,");
      double scg2 = strtod(tok, NULL);
      
      tok = strtok(NULL, "\n\t <>,");
      double scb2 = strtod(tok, NULL);
      
      printf("sphere specular color: %lf, %lf, %lf\n", scr2, scg2, scb2);

      //get sphere coefficients
      fgets(line, 100, file);
      
      //ambient
      tok = strtok(line, "\n\t <>,");
      
      tok = strtok(NULL, "\n\t <>,"); 
      double ac = strtod(tok, NULL);
      
      printf("sphere ambient: %lf\n", ac);
      
      //diffuse
      tok = strtok(NULL, "\n\t <>,"); 
      
      tok = strtok(NULL, "\n\t <>,");
      double dc = strtod(tok, NULL);
      
      printf("sphere diffuse: %lf\n", dc);
      
      //specular
      tok = strtok(NULL, "\n\t <>,"); 
      
      tok = strtok(NULL, "\n\t <>,");
      double sc = strtod(tok, NULL);
      
      printf("sphere specular: %lf\n", sc);
      
      //roughness
      tok = strtok(NULL, "\n\t <>,");
      
      tok = strtok(NULL, "\n\t <>,");
      double ruff = strtod(tok, NULL);

      printf("sphere roughness: %lf\n", ruff);

      //reflection      
      tok = strtok(NULL, "\n\t <>,");
      
      tok = strtok(NULL, "\n\t <>,");
      double refl = strtod(tok, NULL);
      
      printf("sphere reflection: %lf\n", refl);
      
      spheres.push_back(new Sphere(sx, sy, sz, sr, scr, scg, scb, scr2, scg2, scb2,
                                   ac, dc, sc, ruff, refl));
    }
    else if (strncmp(line, "plane", 5) == 0)
    {
      printf("parse plane\n");
      
      //flush out the '{'
      fgets(line, 100, file);
      //get light location
      fgets(line, 100, file);
      
      char *tok = strtok(line, "\n\t <>,");
      double sx = strtod(tok, NULL);
      
      tok = strtok(NULL, "\n\t <>,"); 
      double sy = strtod(tok, NULL);
      
      tok = strtok(NULL, "\n\t <>,");
      double sz = strtod(tok, NULL);
      
      tok = strtok(NULL, "\n\t <>,");
      double sr = strtod(tok, NULL);
      
      printf("sphere position/radius: %lf, %lf, %lf (%lf)\n", sx, sy, sz, sr);
      
      //get sphere diffuse color
      fgets(line, 100, file);
      
      tok = strtok(line, "\n\t <>,"); 
      tok = strtok(NULL, "\n\t <>,");
      
      tok = strtok(NULL, "\n\t <>,");
      double scr = strtod(tok, NULL);
      
      tok = strtok(NULL, "\n\t <>,");
      double scg = strtod(tok, NULL);
      
      tok = strtok(NULL, "\n\t <>,");
      double scb = strtod(tok, NULL);
      
      printf("sphere diffuse color: %lf, %lf, %lf\n", scr, scg, scb);
      
      //get sphere specular color
      fgets(line, 100, file);
      
      tok = strtok(line, "\n\t <>,"); 
      tok = strtok(NULL, "\n\t <>,");
      
      tok = strtok(NULL, "\n\t <>,");
      double scr2 = strtod(tok, NULL);
      
      tok = strtok(NULL, "\n\t <>,");
      double scg2 = strtod(tok, NULL);
      
      tok = strtok(NULL, "\n\t <>,");
      double scb2 = strtod(tok, NULL);
      
      printf("sphere specular color: %lf, %lf, %lf\n", scr2, scg2, scb2);

      //get sphere coefficients
      fgets(line, 100, file);
      
      //ambient
      tok = strtok(line, "\n\t <>,");
      
      tok = strtok(NULL, "\n\t <>,"); 
      double ac = strtod(tok, NULL);
      
      printf("sphere ambient: %lf\n", ac);
      
      //diffuse
      tok = strtok(NULL, "\n\t <>,"); 
      
      tok = strtok(NULL, "\n\t <>,");
      double dc = strtod(tok, NULL);
      
      printf("sphere diffuse: %lf\n", dc);
      
      //specular
      tok = strtok(NULL, "\n\t <>,"); 
      
      tok = strtok(NULL, "\n\t <>,");
      double sc = strtod(tok, NULL);
      
      printf("sphere specular: %lf\n", sc);
      
      //roughness
      tok = strtok(NULL, "\n\t <>,");
      
      tok = strtok(NULL, "\n\t <>,");
      double ruff = strtod(tok, NULL);

      printf("sphere roughness: %lf\n", ruff);

      //reflection      
      tok = strtok(NULL, "\n\t <>,");
      
      tok = strtok(NULL, "\n\t <>,");
      double refl = strtod(tok, NULL);
      
      printf("sphere reflection: %lf\n", refl);
      
      planes.push_back(new Plane(sx, sy, sz, sr, scr, scg, scb, scr2, scg2, scb2,
                                   ac, dc, sc, ruff, refl));
    }
  }
}

int POVRAYParser::finish()
{
  fclose(file);
}
