#include "ray.h"
#include <cmath>

Ray::Ray(double _ox, double _oy, double _oz, double _dx, double _dy, double _dz)
{
  ox = _ox;
  oy = _oy;
  oz = _oz;
  dx = _dx;
  dy = _dy;
  dz = _dz;
  mag = sqrt(dx*dx + dy*dy + dz*dz);
}

void Ray::normalize()
{
  dx /= mag;
  dy /= mag;
  dz /= mag;
  mag = 1.0;
}

double Ray::operator ^(Ray ray)
{
  return dx*ray.dx + dy*ray.dy + dz*ray.dz;
}

double Ray::operator ^(Vertex vert)
{
  return dx*vert.x + dy*vert.y + dz*vert.z;
}

void Ray::operator *(double scalar)
{
  dx = dx*scalar;
  dy = dy*scalar;
  dz = dz*scalar;
}

Ray *Ray::operator %(Ray ray)
{
  double i = dy*ray.dz - dz*ray.dy;
  double j = dz*ray.dx - dx*ray.dz;
  double k = dx*ray.dy - dy*ray.dx;
  
  Ray *r = new Ray(0, 0, 0, i, j, k);
  
  return r;
}

Ray *Ray::operator -(Ray ray)
{
  double i = dx - ray.dx;
  double j = dy - ray.dy;
  double k = dz - ray.dz;
  
  Ray *r = new Ray(0, 0, 0, i, j, k);
  
  return r;
}
  
