#include "raytrace.h"
#include "vertex.h"
#include "ray.h"
#include "pixel.h"
#include "phong.h"
#include "intersection.h"

int last = -1;
int lindex = 0;

int recurse = 0;

Pixel *raytrace(Ray *ray)
{
  Vertex *intersection;
  
  intersection = getIntersections(ray);
  
  if (stages == 0)
  {
    if (last == -1)
    {
      return new Pixel(1, 1, 1);
    }
    else if (last == 0)
    {
      double r = spheres[lindex]->material->diffuse.r;
      double g = spheres[lindex]->material->diffuse.g;
      double b = spheres[lindex]->material->diffuse.b;
    
      return new Pixel(r, g, b);
    }
    else
    {
      double r = planes[lindex]->material->diffuse.r;
      double g = planes[lindex]->material->diffuse.g;
      double b = planes[lindex]->material->diffuse.b;
    
      return new Pixel(r, g, b);
    }    
  }
  else if (stages == 1)
  {
    if (last == -1)
    {
      delete intersection;
      return new Pixel(1, 1, 1);
    }
    else
    {    
      double rgb[3] = {0};
    
      void *obj = 0;
   
      if (last == 0)
        obj = (void*)spheres[lindex];
      else
        obj = (void*)planes[lindex];
    
      getPhongColors(last, obj, ray, rgb, intersection->x, intersection->y, intersection->z);
      
      delete intersection;
      
      return new Pixel(rgb[0], rgb[1], rgb[2]);
    }
  }
  else if (stages == 2)
  {
    if (last == -1)
    {
      delete intersection;
      return new Pixel(1, 1, 1);
    }
    else
    {
      double rgb[3] = {0};
    
      void *obj = 0;
   
      if (last == 0)
        obj = (void*)spheres[lindex];
      else
        obj = (void*)planes[lindex];
    
      getPhongColors_withShadows(last, obj, ray, rgb, intersection->x, intersection->y, intersection->z);
      
      delete intersection;
      
      return new Pixel(rgb[0], rgb[1], rgb[2]);
    }
  }
  else if (stages == 3)
  {
    if (last == -1)
    {
      delete intersection;
      return new Pixel(1, 1, 1);
    }
    else
    {
      static double ref = 0.0;
      
      if (last == 0)
        if  (recurse == 0)
          ref = spheres[lindex]->material->reflectivity;
      
      if (last == 1)
        if  (recurse == 0)
          ref = planes[lindex]->material->reflectivity;
      
      double rgb[3] = {0};
    
      void *obj = 0;
   
      if (last == 0)
        obj = (void*)spheres[lindex];
      else
        obj = (void*)planes[lindex];
    
      getPhongColors_withShadows(last, obj, ray, rgb, intersection->x, intersection->y, intersection->z);
      
      Pixel * p = NULL;
      
      if (recurse < REFLECT_DEPTH)
      {
        Ray incident(intersection->x, intersection->y, intersection->z,
                     ray->ox + intersection->x, ray->oy + intersection->y, ray->oz + intersection->z);
        incident.normalize();
        
        Ray *reflect = NULL;
        
        if (last == 0)
        {
          Sphere *s = static_cast<Sphere*>(obj);
          
          double nx = (intersection->x - s->center->x);
          double ny = (intersection->y - s->center->y);
          double nz = (intersection->z - s->center->z);
      
          Ray snormal(s->center->x, s->center->y, s->center->z, nx, ny, nz);
          snormal.normalize();
   
          double i_dot_n = incident ^ snormal;
          double i_dot_n_2 = i_dot_n * 2;
          
          snormal * i_dot_n_2;
          
          reflect = snormal - incident;
        }
        else
        {
          Plane *p = static_cast<Plane*>(obj);
      
          Ray snormal(p->normal->ox, p->normal->oy, p->normal->oz, p->normal->dx, p->normal->dy, p->normal->dz);
          snormal.normalize();
   
          double i_dot_n = incident ^ snormal;
          double i_dot_n_2 = i_dot_n * 2;
          
          snormal * i_dot_n_2;
          
          reflect = incident - snormal;
        }
        
        recurse++;
        p = raytrace(reflect);
        
        delete reflect;
      }
      
      if (p == NULL)
      {
        return new Pixel(rgb[0], rgb[1], rgb[2]);
      }
      
      double r = rgb[0] + (p->red * ref);
      double g = rgb[1] + (p->green * ref);
      double b = rgb[2] + (p->blue * ref);
      
      recurse--;
      delete p;
      
      return new Pixel(r, g, b);
    }
  }
}
