#include "sphere.h"
#include "vertex.h"
#include "pixel.h"
#include "material.h"

#include <stdlib.h>
#include <stdio.h>
#include <cmath>

Sphere::Sphere(double x, double y, double z, double rad)
{
  center = new Vertex(x, y, z);
  radius = rad;
  material = new Material(1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 1.0, 0);
}

Sphere::Sphere(double x, double y, double z, double rad,
               double dr, double dg, double db,
               double sr, double sg, double sb,
               double ac, double dc, double sc, double ruff, double ref)
{
  center = new Vertex(x, y, z);
  radius = rad;
  material = new Material(dr, dg, db, sr, sg, sb, ac, dc, sc, ruff, ref);
}

Sphere::~Sphere()
{
  delete center;
  delete material;
}

Vertex *Sphere::intersects(Ray *ray)
{
  double a = *ray ^ *ray;
  
  Ray r1(center->x, center->y, center->z, ray->ox - center->x, ray->oy - center->y, ray->oz - center->z);
  
  r1 * 2;
  
  double b = r1 ^ *ray;
  
  r1 * 0.5;
  
  double c = (r1 ^ r1) - (radius*radius);
  
  double discriminant = b*b - 4*a*c;
  
  printf("discriminant: %lf\n", discriminant);
  
  if (discriminant < 0)
  {
    return NULL;
  }

  double t1 = (-b - sqrt(discriminant)) / (2*a);
  double t2 = (-b + sqrt(discriminant)) / (2*a);

  double x1 = ray->ox + (t1 * ray->dx);
  double y1 = ray->oy + (t1 * ray->dy);
  double z1 = ray->oz + (t1 * ray->dz);
  
  double x2 = ray->ox + (t2 * ray->dx);
  double y2 = ray->oy + (t2 * ray->dy);
  double z2 = ray->oz + (t2 * ray->dz);
  
  if (z1 < z2)
    return new Vertex(x1, y1, z1);
  
  return new Vertex(x2, y2, z2);
}
